"""
Routines for multiple scattering. Some modules contain functions to 
explicitly compute the coupling matrix entries. Others contains functions for 
the preparation of lookup tables that are used to approximate the coupling 
matrices by interoplation. The Cython folder has functions to calculate the 
direct coupling block between two particles in the same layer using fast C
routines.
"""
